/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 *------------------------------------------------------*

 File: linfit.c; quick and dirty program for linear fits.
 Usage: linfit [filename] : reads from "filename" (the
 default is stdin) a series of ASCII records until EOF;
 every record should contain a couple of (x,y) values,
 separated from whitespace.  Empty lines are allowed, as
 well as comments started by '#' or '!' and extending to
 end-of-line: comments started by '!' are also printed to
 stdout.  The linear fit results are printed to stdout.

 $Id: linfit.c,v 1.2 2001/04/09 08:56:47 loreti Exp $

 *------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#define BUF_LEN       256
#define COMMENT_KILL  '#'
#define COMMENT_PRINT '!'

void iner(int, char *);

int main(
  int argc,
  char *argv[]
){
  FILE *fp = 0;
  char buffer[BUF_LEN];

  int n         = 0;            /* Number of input data */
  double sum_x  = 0.0;          /* Running sums         */
  double sum_y  = 0.0;
  double sum_x2 = 0.0;
  double sum_y2 = 0.0;
  double sum_xy = 0.0;
  double delta, a, b;           /* Linear fit, y = a + b x        */
  double mean_x, mean_y;        /* Mean values as samples         */
  double var_x, var_y;          /* Variances as samples           */
  double cov_xy;                /* Covariance as samples          */
  double sumreq = 0.0;          /* Sum of squared residuals       */
  double r      = 0.0;          /* Linear correlation coefficient */
  double eqm_y  = 0.0;          /* Estimated errors (RMS) for y,  */
  double eqm_a  = 0.0;          /*   a,                           */
  double eqm_b  = 0.0;          /*     and b.                     */
  double mean_r, eqm_r;         /* Mean number of runs, and EQM   */
  int curs      = 0;            /* Current residual sign          */
  int nRuns     = 0;            /* Number of runs                 */
  int nResPos   = 0;            /* Number of positive             */
  int nResNeg   = 0;            /*   and of negative residuals.   */
  int lineNo    = 0;            /* Line currently being read      */

/**
 | Input file handling (the file will be silently closed by the
 | operating system when exiting).
**/

  if (argc > 2) {
    fprintf(stderr, "Usage: %s [filename]\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  if (argc == 2) {
    if ((fp = fopen(argv[1], "r")) == 0) {
      fprintf(stderr, "Error opening file \"%s\"\n", argv[1]);
      exit(EXIT_FAILURE);
    }
  } else {
    fp = stdin;
  }

/**
 | Data input
**/

  while (fgets(buffer, BUF_LEN, fp) != 0) {
    double x, y;
    char *pc;

    ++lineNo;
    if ((pc = strchr(buffer, COMMENT_KILL))  != 0) *pc = '\0';
    if ((pc = strchr(buffer, COMMENT_PRINT)) != 0) {
      printf(pc);
      *pc = '\0';
    }

    for (pc = buffer;  *pc != '\0' && isspace((unsigned char)*pc);  ++pc) { }
    if (*pc != '\0') {
      ++n;
      x = strtod(pc, &pc);
      if (isspace((unsigned char) *pc)) {
        y = strtod(pc, &pc);
        if (isspace((unsigned char) *pc)  ||  *pc == '\0') {
          sum_x  += x;
          sum_y  += y;
          sum_x2 += x * x;
          sum_y2 += y * y;
          sum_xy += x * y;
        } else {
          iner(lineNo, buffer);
        }
      } else {
        iner(lineNo, buffer);
      }
    }
  }

  if (ferror(fp)) {
    fprintf(stderr, "Error reading file \"%s\"\n", argv[1]);
    perror("fgets");
    exit(EXIT_FAILURE);
  }

/**
 | Linear fit proper
**/

  if (n < 2) {
    fputs("You need at least 2 data for a linear fit!\n", stderr);
    exit(EXIT_FAILURE);
  }

  delta = n * sum_x2 - sum_x * sum_x;
  a     = (sum_x2 * sum_y - sum_x * sum_xy) / delta;
  b     = (n * sum_xy - sum_x * sum_y) / delta;

  if (n > 2) {
    mean_x = sum_x / n;
    mean_y = sum_y / n;
    cov_xy = sum_xy / n - mean_x * mean_y;
    var_x  = sum_x2 / n - mean_x * mean_x;
    var_y  = sum_y2 / n - mean_y * mean_y;
    r      = cov_xy / sqrt(var_x * var_y);
    sumreq = n * var_y * (1 - r * r);
    eqm_y  = sqrt(sumreq / (n - 2));
    eqm_a  = eqm_y * sqrt(sum_x2 / delta);
    eqm_b  = eqm_y * sqrt(n / delta);
  }

/**
 | Printout (part 1)
**/

  puts("--------------- Unweighted linear fit (y = A + B x) ---------------");
  printf("N    = %d\n", n);
  printf("yRMS = %.6g\n", eqm_y);
  printf("r    = %.6g\n", r);
  printf("A    = %.6g +- %.6g\n", a, eqm_a);
  printf("B    = %.6g +- %.6g\n", b, eqm_b);
  puts("---------------------------------------------");
  puts("              X              Y       Residual");

/**
 | Returns to the beginning of the file; reads a second time the input
 | data, to compute residuals.
**/

  rewind(fp);
  while (fgets(buffer, BUF_LEN, fp) != 0) {
    char *pc;

    if ((pc = strchr(buffer, COMMENT_KILL))  != 0) *pc = '\0';
    if ((pc = strchr(buffer, COMMENT_PRINT)) != 0) *pc = '\0';

    for (pc = buffer;  *pc != '\0' && isspace((unsigned char)*pc);  ++pc) { }
    if (*pc != '\0') {
      double x, y, resid;

      x = strtod(pc, &pc);
      y = strtod(pc, &pc);
      resid = y - (a + b * x);

      if (resid >= 0.0) {
        nResPos++;
        if (curs != 1) {
          curs = 1;
          nRuns++;
        }
      } else {
        nResNeg++;
        if (curs != -1) {
          curs = -1;
          nRuns++;
        }
      }

      printf("%15.6g%15.6g%15.6g\n", x, y, resid);
    }
  }

  mean_r  = 2.0 * nResPos * nResNeg;
  eqm_r   = mean_r /= n;
  eqm_r  *= (2.0 * nResPos * nResNeg - n);
  eqm_r  /= n;
  eqm_r  /= (n - 1);
  eqm_r   = sqrt(eqm_r);
  mean_r += 1.0;

  puts("---------------------------------------------");
  printf("%d runs (expected %.6g +- %.6g)\n", nRuns, mean_r, eqm_r);
  printf("Sum of squared residuals: %.6g\n", sumreq);

  return EXIT_SUCCESS;
}

void iner(
  int n,
  char *pc
){
  fprintf(stderr, "Input error in record nr. %d: %s", n, pc);
  exit(EXIT_FAILURE);
}
