//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// File: linfit.cxx; quick and dirty program for linear fits.
//
// Usage: linfit [filename]
//
// linfxx reads from "filename" (the default is the standard input
// stream) a series of ASCII records until the end-of-file; every
// record should contain a couple of (x,y) values, separated from
// whitespace.  Empty lines are allowed, as well as comments started
// by "#" or by "!" and ended by the end-of-line; comments started by
// "!" are, in addition, printed on the standard output stream.  The
// results of the unweighted linear fit are also printed to the
// standard output stream.
//
// $Id: linfit.cxx,v 1.12 2006/06/12 06:22:33 loreti Exp $
//
// -------------------------------------------------------------------

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>

#include <cctype>
#include <cmath>
#include <cstdlib>

using namespace std;

// g++ version 2 has a Standard Library implementation that is non
// completely conforming with reference to iostreams; right() is not
// correctly defined.  According to ISO/IEC 14882:1998 27.4.5.2,
// std::ios_base::right(std::ios_base & str) is supposed to call
// str.setf(ios_base::right, ios_base::adjustfield) and return str.

#if __GNUG__ < 3
inline ostream & right (ostream & i) { i.setf(i.right, i.adjustfield);
  return i; }
#endif // g++ version 2

// A "point" to be fitted is abstracted by class Point as a couple of
// coordinates; from Point it is easy to publicly inherit in order to
// define points having different errors on the y-coordinate.
//
// Supplied methods:
// - ctors: default; from numerical values; and from a string to be
//   interpreted (throws for invalid input)
// - dtor
// - inspectors: to get x and y
// - a comparison operator: less, needed for sorting;
// - a method to output the coordinates on an output stream.

class Point {
private:
  double x_;
  double y_;

public:
  Point(double x = 0.0, double y = 0.0) : x_(x), y_(y) { }
  Point(const string &, const size_t);

  ~Point() { }

  double getx() const { return x_; }
  double gety() const { return y_; }

  bool operator < (const Point & rhs) const {
    return (x_ < rhs.x_   ||   (x_ == rhs.x_  &&  y_ < rhs.y_));
  }

  void out(ostream & o = cout) const {
    o << '(' << x_ << ", " << y_ << ')'; }
};

// Methods for class Point

Point::Point(
  const string & s,
  const size_t   line
) {
  char *endPtr;

  x_ = strtod(s.c_str(), &endPtr);
  if (!isspace(static_cast<unsigned char>(*endPtr))) throw line;
  y_ = strtod(endPtr, &endPtr);
  if (!isspace(static_cast<unsigned char>(*endPtr))  &&
      *endPtr != '\0') throw line;
}

// A procedure Fit, that from a vector of Points implements the least
// squares fitting procedure and prints the fit results and the
// residual analysis.

void Fit(
  const vector<Point> & v,
  ostream &             o = cout
) {
  if (v.size() < 2) {
    throw runtime_error("Less than 2 points in input");
  }

  vector<Point> * vLoc;         // Points to be fitted (local copy)
  double          sum_x  = 0.0; // Running sums
  double          sum_y  = 0.0;
  double          sum_x2 = 0.0;
  double          sum_y2 = 0.0;
  double          sum_xy = 0.0;

  vLoc = new vector<Point>( v );
  sort(vLoc->begin(), vLoc->end());

  vector<Point>::const_iterator       it;
  const vector<Point>::size_type      n    = vLoc->size();
  const vector<Point>::const_iterator iEnd = vLoc->end();

  for (it = vLoc->begin();   it != iEnd;   ++it) {
    register double x = it->getx();
    register double y = it->gety();

    sum_x  += x;
    sum_y  += y;
    sum_x2 += x * x;
    sum_y2 += y * y;
    sum_xy += x * y;
  }

  double a;                     // Coefficients:
  double b;                     //   (y = A + B * x).
  double r      = 0.0;          // Linear correlation coefficient
  double eqm_y  = 0.0;          // Standard deviation for Y's,
  double eqm_a  = 0.0;          //   and for the resulting values
  double eqm_b  = 0.0;          //     of A and B.
  double sumreq = 0.0;          // Sum of squared residuals
  double delta;
  double mean_x;
  double mean_y;
  double var_x;
  double var_y;
  double cov_xy;

  delta  = n * sum_x2 - sum_x * sum_x;
  a      = (sum_x2 * sum_y - sum_x * sum_xy) / delta;
  b      = (n * sum_xy - sum_x * sum_y) / delta;
  mean_x = sum_x / n;
  mean_y = sum_y / n;
  cov_xy = sum_xy / n - mean_x * mean_y;
  var_x  = sum_x2 / n - mean_x * mean_x;
  var_y  = sum_y2 / n - mean_y * mean_y;

  if (n > 2) {
    r      = cov_xy / sqrt(var_x * var_y);
    sumreq = n * var_y * (1 - r * r);
    eqm_y  = sqrt(sumreq / (n - 2));
    eqm_a  = eqm_y * sqrt(sum_x2 / delta);
    eqm_b  = eqm_y * sqrt(n / delta);
  }

  // Fit results

  o << "--------------- Unweighted linear fit (y = A + B x) ---------------\n";
  o << "N    = " << vLoc->size() << '\n';
  o << "yRMS = " << eqm_y        << '\n';
  o << "r    = " << r            << '\n';
  o << "A    = " << a            << " +- " << eqm_a << '\n';
  o << "B    = " << b            << " +- " << eqm_b << endl;

  o << "---------------------------------------------\n";
  o.width(15);  o << right << "X";
  o.width(15);  o << right << "Y";
  o.width(15);  o << right << "Residual" << '\n';

  vector<Point>::size_type nPos = 0;
  vector<Point>::size_type nNeg = 0;
  vector<Point>::size_type nRun = 0;
  int                      curs = 0;

  for (it = vLoc->begin();   it != iEnd;   ++it) {
    register double x   = it->getx();
    register double y   = it->gety();
             double res = y - (a + b * x);

    o.width(15);  o << x;
    o.width(15);  o << y;
    o.width(15);  o << res << '\n';

    if (res >= 0.0) {
      ++nPos;
      if (curs != 1) {
        curs = 1;
        ++nRun;
      }
    } else {
      ++nNeg;
      if (curs != -1) {
        curs = -1;
        ++nRun;
      }
    }
  }
  o << "---------------------------------------------\n";

  // Residual analysis

  double runs, runm;

  runm  = 2.0 * nPos * nNeg;
  runs  = runm /= n;
  runs *= (2.0 * nPos * nNeg - n);
  runs /= n;
  runs /= (n - 1);
  runs  = sqrt(runs);
  runm += 1.0;

  o << nRun << " run" << (nRun == 1 ? "" : "s")
    << " (expected " << runm << " +- " << runs << ")\n";
  o << "Sum of squared residuals: " << sumreq << endl;

  delete vLoc;
}

// Main program.  Tests the supplied argument, reads, fits and prints
// the results on stdout.

int main(
  int   argc,
  char *argv[]
) {
  if (argc > 2) {
    cerr << "Usage:\t" << argv[0] << " [filename]\n";
    cerr << "\t'filename' defaults to the standard input stream\n";
    return EXIT_FAILURE;
  }

  ifstream inFile;

  if (argc == 2) {

    // The input file name has been explicitely given.  Opens the file
    // and, if the operation is successfull, links std::cin to that
    // file.

    inFile.open(argv[1]);
    if (!inFile) {
      cerr << "Couldn't open input file \"" << argv[1] << "\"\n";
      return EXIT_FAILURE;
    }
    cin.rdbuf( inFile.rdbuf() );
  }

  Point         p;
  vector<Point> v;
  string        s;
  size_t        lineNo = 0;

  while (getline(cin, s)) {
    ++lineNo;

    // Remove comments, and skip empty lines.
    //
    // A comment is started by a '#' or a '!' and terminated by the
    // end of the line; comments started by '!' are also printed on
    // the standard output stream.

    string::size_type comment = s.find_first_of("#!");
    if (comment != string::npos) {
      if (s[comment] == '!') cout << s.substr(comment) << endl;
      s.erase(comment);
    }

    if (s.find_first_not_of(" \t") != string::npos) {
      try {
        p = Point(s, lineNo);
      } catch (const size_t i) {
        cerr << "Error: invalid input in line " << i << '\n';
        return EXIT_FAILURE;
      }
      v.push_back(p);
    }
  }

  try {
    Fit(v);
  } catch (const exception & e) {
    cerr << "Error: " << e.what() << '\n';
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}
