//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// File: linfxw.cxx; quick and dirty program for weighted linear fits.
//
// Usage: linfxw [filename]
//
// linfxw reads from "filename" (the default is the standard input
// stream) a series of ASCII records until the end-of-file; every
// record should contain a set of (x,y,s) values, separated from
// whitespace ("s" being the standard deviation of "y").  Empty lines
// are allowed, as well as comments started by "#" or by "!" and ended
// by the end-of-line; comments started by "!" are, in addition,
// printed on the standard output stream.  The results of the weighted
// linear fit are also printed to the standard output stream.
//
// $Id: linfxw.cxx,v 1.10 2006/06/12 06:22:33 loreti Exp $
//
// -------------------------------------------------------------------

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>

#include <cctype>
#include <cmath>
#include <cstdlib>

using namespace std;

// g++ version 2 has a Standard Library implementation that is non
// completely conforming with reference to iostreams; right() is not
// correctly defined.  According to ISO/IEC 14882:1998 27.4.5.2,
// std::ios_base::right(std::ios_base & str) is supposed to call
// str.setf(ios_base::right, ios_base::adjustfield) and return str.

#if __GNUG__ < 3
inline ostream & right (ostream & i) { i.setf(i.right, i.adjustfield);
  return i; }
#endif // g++ version 2

// A "point" to be fitted is abstracted by class Point as a couple of
// coordinates and a standard deviation.
//
// Supplied methods:
// - ctors: default; from numerical values; and from a string to be
//   interpreted (throws for invalid input)
// - dtor
// - inspectors: to get x, y and s;
// - a comparison operator: less, needed for sorting;
// - a method to output the coordinates on an output stream.

class Point {
private:
  double x_;
  double y_;
  double s_;

public:
  Point(double x = 0.0, double y = 0.0, double s = 1.0) : x_(x), y_(y), s_(s)
    { }
  Point(const string &, const int);

  ~Point() { }

  double getx() const { return x_; }
  double gety() const { return y_; }
  double gets() const { return s_; }

  bool operator < (const Point & rhs) const {
    return (x_ < rhs.x_   ||   (x_ == rhs.x_  &&  y_ < rhs.y_));
  }

  void out(ostream & o = cout) const {
    o << '(' << x_ << ", " << y_ << " +- " << s_ << ')';
  }
};

// Methods for class Point

Point::Point(
  const string & s,
  const int      line
) {
  char *endPtr;

  x_ = strtod(s.c_str(), &endPtr);
  if (!isspace(static_cast<unsigned char>(*endPtr))) throw line;
  y_ = strtod(endPtr, &endPtr);
  if (!isspace(static_cast<unsigned char>(*endPtr))) throw line;
  s_ = strtod(endPtr, &endPtr);
  if (!isspace(static_cast<unsigned char>(*endPtr))  &&  *endPtr != '\0')
    throw  line;
  if (s_ <= 0.0) throw -line;
}

// A procedure Fit, that from a vector of Points implements the least
// squares fitting procedure and prints the fit results and the
// residual analysis.

void Fit(
  const vector<Point> & v,
  ostream &             o = cout
) {
  if (v.size() < 2) {
    throw runtime_error("Less than 2 points in input");
  }

  vector<Point> * vL;           // Points to be fitted (local copy)

  vL = new vector<Point>( v );
  sort(vL->begin(), vL->end());

  double sum_x  = 0.0;          // Running sums
  double sum_y  = 0.0;
  double sum_x2 = 0.0;
  double sum_y2 = 0.0;
  double sum_xy = 0.0;
  double sum_w  = 0.0;
  double mean_x = 0.0;
  double mean_y = 0.0;
  double var_x  = 0.0;
  double var_y  = 0.0;
  double cov_xy = 0.0;

  vector<Point>::const_iterator       it;
  const vector<Point>::size_type      n    = vL->size();
  const vector<Point>::const_iterator iEnd = vL->end();

  for (it = vL->begin();   it != iEnd;   ++it) {
    double x = it->getx();
    double y = it->gety();
    double w = it->gets();

    w      = 1.0 / (w * w);
    sum_w += w;

    double xw = x * w;
    double yw = y * w;

    sum_x  +=     xw;
    sum_y  +=     yw;
    sum_x2 += x * xw;
    sum_y2 += y * yw;
    sum_xy += x * yw;

    mean_x += x;
    mean_y += y;
    var_x  += x * x;
    var_y  += y * y;
    cov_xy += x * y;
  }

  double a;                     // Coefficients:
  double b;                     //   (y = A + B * x).
  double r     = 0.0;           // Linear correlation coefficient
  double eqm_a = 0.0;           // Standard deviation for the
  double eqm_b = 0.0;           //   resulting values of A and B.
  double delta;

  delta  =  sum_w  * sum_x2 - sum_x * sum_x;
  a      = (sum_x2 * sum_y  - sum_x * sum_xy) / delta;
  b      = (sum_w  * sum_xy - sum_x * sum_y)  / delta;

  mean_x = mean_x / n;
  mean_y = mean_y / n;
  cov_xy = cov_xy / n - mean_x * mean_y;
  var_x  = var_x  / n - mean_x * mean_x;
  var_y  = var_y  / n - mean_y * mean_y;

  if (n > 2) {
    r      = cov_xy / sqrt(var_x * var_y);
    eqm_a  = sqrt(sum_x2 / delta);
    eqm_b  = sqrt(sum_w  / delta);
  }

  // Fit results

  o << "---------------- Weighted linear fit (y = A + B x) ----------------\n";
  o << "N = " << vL->size() << '\n';
  o << "r = " << r          << '\n';
  o << "A = " << a          << " +- " << eqm_a << '\n';
  o << "B = " << b          << " +- " << eqm_b << endl;

  // Residual analysis

  o << "------------------------------------------------------------\n";
  o.width(15);  o << right << "X";
  o.width(15);  o << right << "Y";
  o.width(15);  o << right << "sigma(Y)";
  o.width(15);  o << right << "Residual" << '\n';

  vector<Point>::size_type nPos = 0;
  vector<Point>::size_type nNeg = 0;
  vector<Point>::size_type nRun = 0;
  int                      curs = 0;
  double                   chi2 = 0.0;

  for (it = vL->begin();   it != iEnd;   ++it) {
    double x   = it->getx();
    double y   = it->gety();
    double s   = it->gets();
    double res = y - (a + b * x);
    double rw  = res / s;

    chi2 += rw * rw;

    o.width(15);  o << x;
    o.width(15);  o << y;
    o.width(15);  o << s;
    o.width(15);  o << res << '\n';

    if (res >= 0.0) {
      ++nPos;
      if (curs != 1) {
        curs = 1;
        ++nRun;
      }
    } else {
      ++nNeg;
      if (curs != -1) {
        curs = -1;
        ++nRun;
      }
    }
  }
  o << "------------------------------------------------------------\n";

  double runs, runm;

  runm  = 2.0 * nPos * nNeg;
  runs  = runm /= n;
  runs *= (2.0 * nPos * nNeg - n);
  runs /= n;
  runs /= (n - 1);
  runs  = sqrt(runs);
  runm += 1.0;

  o << nRun << " run" << (nRun == 1 ? "" : "s")
    << " (expected " << runm << " +- " << runs << ")\n";
  o << "Chi square: " << chi2 << ", with "
    << n - 2 << " degree" << ((n - 2) == 1 ? "" : "s")
    << " of freedom\n";

  delete vL;
}

// Main program.  Tests the supplied argument, reads, fits and prints
// the results on stdout.

int main(
  int   argc,
  char *argv[]
) {
  if (argc > 2) {
    cerr << "Usage:\t" << argv[0] << " [filename]\n";
    cerr << "\t'filename' defaults to the standard input stream\n";
    return EXIT_FAILURE;
  }

  ifstream inFile;

  if (argc == 2) {

    // The input file name has been explicitely given.  Opens the file
    // and, if the operation is successfull, links std::cin to that
    // file.

    inFile.open(argv[1]);
    if (!inFile) {
      cerr << "Couldn't open input file \"" << argv[1] << "\"\n";
      return EXIT_FAILURE;
    }
    cin.rdbuf( inFile.rdbuf() );
  }

  Point         p;
  vector<Point> v;
  string        s;
  int           lineNo = 0;

  while (getline(cin, s)) {
    ++lineNo;

    // Remove comments, and skip empty lines
    //
    // A comment is started by a '#' or a '!' and terminated by the
    // end of the line; comments started by '!' are also printed on
    // the standard output stream.

    string::size_type comment = s.find_first_of("#!");
    if (comment != string::npos) {
      if (s[comment] == '!') cout << s.substr(comment) << endl;
      s.erase(comment);
    }

    if (s.find_first_not_of(" \t") != string::npos) {
      try {
        p = Point(s, lineNo);
      } catch (const int i) {
        if (i > 0) {
          cerr << "Error: invalid input in line " <<  i << '\n';
        } else {
          cerr << "Error: invalid error in line " << -i << '\n';
        }
        return EXIT_FAILURE;
      }
      v.push_back(p);
    }
  }

  try {
    Fit(v);
  } catch (const exception & e) {
    cerr << "Error: " << e.what() << '\n';
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}
