/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@pd.infn.it                            |
 | WWW:    http://www.pd.infn.it/~loreti/mlo.html       |
 *------------------------------------------------------*

 This program computes the probability of the residual runs, from
 the number of positive (N+) and negative (N-) residuals.  A "run"
 is a sequence of residuals having the same sign; the probability
 for a number of runs R may be computed using formulas given
 e.g. in:

 - Eadie, Drijard, James, Roos, Sadoulet - Statistical methods in
   experimental physics - North-Holland 1988 - Pag. 263

 - Barlow - Statistics - Wiley 1997 - Pag. 154

 $Id: run.c,v 1.6 2006/09/11 12:48:45 loreti Exp $

 N.B.: C-u 66 M-x set-fill-column

 *--------------------------------------------------------------*/

/**
 | Include files
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

/**
 | A macro to test if strtoul has stopped on an invalid numeric
 | character
**/

#define LAST_FIELD (isspace((unsigned char) *pc)  ||  *pc == '\0')

/**
 | Global variables: the executable name
**/

char *progName;

char *baseName(char *);
void  printRun(unsigned int, unsigned int);
void  usage(void);

int main(
  int   argc,
  char *argv[]
) {

  /**
   | The main program: checks the command line parameters, then
   | calls "printRun" to do the job.
  **/

  unsigned int  nplus, nminus;
  char         *pc;

  progName = baseName(argv[0]);

  if (argc != 3) {
    usage();
    return EXIT_FAILURE;
  }

  nplus = strtoul(argv[1], &pc, 0);
  if (!LAST_FIELD) {
    usage();
    fprintf(stderr, "==>\tInvalid character in N+ = \"%s\"\n",
            argv[1]);
    return EXIT_FAILURE;
  }

  nminus = strtoul(argv[2], &pc, 0);
  if (!LAST_FIELD) {
    usage();
    fprintf(stderr, "==>\tInvalid character in N- = \"%s\"\n",
            argv[2]);
    return EXIT_FAILURE;
  }

  if (nplus > 0   &&   nminus > 0) {
    printRun(nplus, nminus);
  } else {
    usage();
    fputs("==>\tN+ and N+ must be at least 1\n", stderr);
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}

char *baseName(
  char *pc
){

  /**
   | Strips the (eventual) path informations from *pc
  **/

  char *p;

  p = strrchr(pc, '/');
  if (p == 0) return pc;
  return ++p;
}

void printRun(
  unsigned int n1,      /* Positive residuals (points "above") */
  unsigned int n2       /* Negative residuals (points "below") */
) {

  /**
   | Local variables:
   | i:       generic variable;
   | n:       total number of points (above and below);
   | m:       the smallest between n1 and n2;
   | rmax:    the maximum possible number of runs;
   | r:       the number of runs being currently considered;
   | prob:    the probability of a given r;
   | denom:   the denominator of prob, i.e. the binomial
   |          coefficient \binom{n}{n1} = \binom{n}{n2} ;
   | numEven: the numerator of prob, when r is even;
   | numOdd:  the numerator of prob, when r is odd;
   | sum:     the cumulative probability of all r's seen so far;
   | mean:    the mean value E(r), that may be computed using only
   |          n1 and n2;
   | var:     the variance Var(r).
   |
   | Some variables are declared double to avoid overflows, also
   | if their content is intrinsically integer.
  **/

  unsigned int i;
  unsigned int n = n1 + n2;
  unsigned int m;
  unsigned int rmax;
  unsigned int r;

  double prob;
  double denom   = 1.0;
  double numEven = 2.0;
  double numOdd  = n - 2;
  double sum     = 0.0;
  double mean;
  double var;

  /**
   | Temporary variables used to store various numeric factors
   | while iterating over r
  **/

  unsigned int sEven, sOdd, f1, f2, g1, g2;

  /**
   | E(r) and Var(r) may be printed immediately
  **/

  mean = 1.0 + 2.0 * n1 * n2 / n;
  var  = 2.0 * n1 * n2 * (2.0 * n1 * n2 - n) / (n * n * (n - 1));
  printf("E(R) = %f, Var(R) = %f, EQM(R) = %f\n",
         mean, var, sqrt(var));

  /**
   | Computes rmax: the maximum possible number of runs is n1+n2
   | when n1 and n2 are the same (the signs may alternate); or is
   | 2*min(n1,n2)+1 otherwise (when all the residuals having the
   | sign less frequently seen are intermixed with one or more
   | values of the opposite sign).
  **/

  if (n1 == n2) {
    m    = n1;
    rmax = n;
  } else {
    m    = n1 < n2 ? n1 : n2;
    rmax = 2 * m + 1;
  }

  /**
   | Stores in denom the binomial coefficient \binom{n}{m},
   | computed recursively from:
   | (1a)         \binom{n}{0} = 1
   | (1b)         \binom{n}{k} = \binom{n}{k-1} \frac{n-k+1}{k}
  **/

  for (r = 0,  i = n;   r < m;   ) {
    denom *= i--;
    denom /= ++r;
  }

  /**
   | For all possible r's from 2 to rmax computes the probability
   | as ratio between a numerator (different for even and odd
   | terms) and the common denominator denom.  The numerators
   | (numOdd and numEven) are computed recursively using (1b)
   | above, as:
   |
   | (2a) If R is even, being R = 2s:
   | (2b) when s = 1, numEven(1) = 2;
   | (2c) numEven(s+1) = numEven(s) \frac{(n1-s)(n2-s)}{s^2} .
   |
   | (3a) If R is odd, being R = 2s-1:
   | (3b) when s=2, numOdd(2) = n-2;
   | (3c) numOdd(s+1) = numOdd(s) \frac{(n1-s+1)(n2-s+1}{(n-2s+2)}
   |                    \frac{n-2s}{s(s-1)}
  **/

  sOdd  = 2;
  sEven = 1;
  f1    = g1 = n1 - 1;
  f2    = g2 = n2 - 1;

  for (r = 2;   r <= rmax;   r++) {
    if (r & 1) {
      prob    = numOdd;         /* r odd */
      numOdd *= f1;
      numOdd *= f2;
      numOdd /= f1-- + f2--;
      numOdd *= n - sOdd - sOdd;
      numOdd /= sOdd - 1;
      numOdd /= sOdd++;
    } else {
      prob     = numEven;       /* r even */
      numEven *= g1--;
      numEven /= sEven;
      numEven *= g2--;
      numEven /= sEven++;
    }
    prob /= denom;
    sum  += prob;
    printf("R = %d\tP = %11.5g (sum = %11.5g)\n", r, prob, sum);
  }
}

void usage()
{
  printf("Usage:\t%s N+ N-\n", progName);
  puts("==>\tGiven the number of points with positive "
       "(N+) and negative");
  puts("\t(N-) residuals, the statistics of residual runs "
       "is printed.");
}
