/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

                      File: statfun.c

 $Id: statfun.c,v 1.2 2004/10/07 09:00:20 loreti Exp $

 *------------------------------------------------------------------*

 Tutte le funzioni necessarie al calcolo delle grandezze statistiche
 considerate in "statmain" sono incluse nella GNU scientific library
 (GSL), con la sola eccezione della distribuzione di probabilita' per
 il test di Kolmogorov e Smirnov; comunque in questo file ci sono
 tutte le procedure necessarie, in modo da rendere l'uso del programma
 indipendente dalla preventiva installazione della GSL.

                     Ripreso con cambiamenti da:
               Press, Teukolsky, Vetterling e Flannery
              Numerical recipes in C (Seconda edizione)
                   Cambridge University Press, 1992

 *------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "statfun.h"

#ifndef HAVE_GSL

static double betacf(double, double, double);
static double gammln(double);
static void   gcf(double *, double, double, double *);
static void   gser(double *, double, double, double *);

double betai(
  double a,
  double b,
  double x
){

  /**
   | Funzione Beta Incompleta  I_x(a,b)
  **/

  double bt;

  if (x < 0.0   ||   x > 1.0) {
    puts("Valore illegale di x in \"betai\"");
    exit(EXIT_FAILURE);
  }
  if (x == 0.0   ||   x == 1.0) {
    bt = 0.0;
  } else {
    bt = exp(gammln(a + b) - gammln(a) -gammln(b) +
         a * log(x) + b * log(1.0 - x));
  }
  if (x   <   (a + 1.0) / (a + b + 2.0)) {
    return (bt * betacf(a, b, x) / a);
  } else {
    return (1.0 - bt * betacf(b, a, 1.0 - x) / b);
  }
}

double erf(
  double x
){

  /**
   | Gauss error function
  **/

  return (x < 0.0) ? (-gammp(0.5, x*x)) : (gammp(0.5, x*x));
}

double gammp(
  double a,
  double x
){

  /**
   | Funzione Gamma Incompleta   -   P(a,x)
  **/

  double gamser, gammcf, gln;

  if (x < 0.0   ||   a <= 0.0) {
    puts("Argomenti illegali nella routine \"gammp\"");
    exit(EXIT_FAILURE);
  }
  if (x < (a + 1.0)) {
    gser(&gamser, a, x, &gln);
    return gamser;
  } else {
    gcf(&gammcf, a, x, &gln);
    return (1.0 - gammcf);
  }
}

double gammq(
  double a,
  double x
){

  /**
   | Funzione Gamma Incompleta   -   Q(a,x) = 1 - P(a,x)
  **/

  double gamser, gammcf, gln;

  if (x < 0.0   ||   a <= 0.0) {
    puts("Argomenti illegali nella routine \"gammq\"");
    exit(EXIT_FAILURE);
  }

  if (x < (a + 1.0)) {
    gser(&gamser, a, x, &gln);
    return (1.0 - gamser);
  } else {
    gcf(&gammcf, a, x, &gln);
    return gammcf;
  }
}

static double betacf(
  double a,
  double b,
  double x
){

  /**
   | Espansione in frazioni continue della Funzione Beta Incompleta
  **/

  int    m, m2;
  double aa, c, d, del, h, qab, qam, qap;

  qab = a + b;
  qap = a + 1.0;
  qam = a - 1.0;
  c   = 1.0;
  d   = 1.0 - qab * x / qap;
  if (fabs(d) < FPMIN) d = FPMIN;

  d = 1.0 / d;
  h = d;
  for (m = 1;  m <= MAXIT;  m++) {
    m2 = m + m;
    aa = m * (b - m) * x / ((qam + m2) * (a + m2));
    d  = 1.0 + aa * d;
    if (fabs(d) < FPMIN) d = FPMIN;

    c = 1.0 + aa / c;
    if (fabs(c) < FPMIN) c = FPMIN;

    d   = 1.0 / d;
    h  *= (d * c);
    aa  = -(a + m) * (qab + m) * x / ((a + m2) * (qap + m2));
    d   = 1.0 + aa * d;
    if (fabs(d) < FPMIN) d = FPMIN;

    c = 1.0 + aa / c;
    if (fabs(c) < FPMIN) c = FPMIN;

    d    = 1.0 / d;
    del  = d * c;
    h   *= del;
    if (fabs(del - 1.0) < EPS) break;
  }
  if (m > MAXIT) {
    puts("a o b troppo grande o MAXIT troppo piccolo in \"betacf\"");
    exit(EXIT_FAILURE);
  }
  return h;
}

static double gammln(
  double xx
){

  /**
   | Logaritmo della Funzione Gamma
  **/

  double x, y, tmp, ser;
  int j;
  static double cof[6] = {
    76.18009172947146, -86.50532032941677, 24.01409824083091,
    -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5
  };

  y    = x = xx;
  tmp  = x + 5.5;
  tmp -= (x + 0.5) * log(tmp);
  ser  = 1.000000000190015;
  for (j = 0;  j < 6;  j++) {
    y   += 1.0;
    ser += cof[j] / y;
  }
  return -tmp + log(2.5066282746310005 * ser / x);
}

static void gcf(
  double *gammcf,
  double  a,
  double  x,
  double *gln
){

  /**
   | Calcolo delle Funzione Gamma e della Funzione Gamma Incompleta
   | Q(a,x) dallo sviluppo in frazione continua.
  **/

  int i;
  double an, b, c, d, del, h;

  *gln = gammln(a);
  b    = x + 1.0 - a;
  c    = 1.0 / FPMIN;
  h    = d = 1.0 / b;
  for (i = 1;  i <= ITMAX;  i++) {
    an  = -i * (i - a);
    b  += 2.0;
    d   = an * d + b;
    if (fabs(d) < FPMIN) d = FPMIN;

    c = b + an / c;
    if (fabs(c) < FPMIN) c = FPMIN;

    d    = 1.0 / d;
    del  = d * c;
    h   *= del;
    if (fabs(del - 1.0) < EPS) break;
  }
  if (i > ITMAX ) {
    puts("a troppo grande o ITMAX troppo piccolo in \"gcf\"");
    exit(EXIT_FAILURE);
  }
  *gammcf = exp(-x + a * log(x) - (*gln)) * h;
}

static void gser(
  double *gamser,
  double  a,
  double  x,
  double *gln
){

  /**
   | Calcolo delle Funzione Gamma e della Funzione Gamma Incompleta
   | P(a,x) dallo sviluppo in serie.
  **/

  int    n;
  double sum, del, ap;

  *gln = gammln(a);
  if (x <= 0.0) {
    if (x < 0.0) {
      puts("x<0 nella routine \"gser\"");
      exit(EXIT_FAILURE);
    }
    *gamser = 0.0;
    return;
  } else {
    ap  = a;
    del = sum = 1.0 / a;

    for (n = 1;   n <= ITMAX;   n++) {
      ap  += 1.0;
      del *= (x / ap);
      sum += del;

      if (fabs(del) < (fabs(sum) * EPS)) {
        *gamser = sum * exp(-x + a * log(x) - (*gln));
        return;
      }
    }
    puts("a troppo grande o ITMAX troppo piccolo in \"gcf\"");
    exit(EXIT_FAILURE);
  }
}

#endif /* HAVE_GSL */

double probks(
  double alam
){
/**
 | Funzione di probabilita' di Kolmogorov & Smirnov
**/

  int    j;
  double a2, term, temp;
  double fac    = 2.0;
  double sum    = 0.0;
  double termbf = 0.0;

  a2 = -2.0 * alam * alam;
  for (j = 1;   j <= KS_ITER;   j++) {
    term  = fac * exp(a2 * j * j);
    sum  += term;
    temp  = fabs(term);

    if (temp <= KS_EPS1 * termbf   ||   temp <= KS_EPS2 * sum) {
      return sum;
    }
    fac    = -fac;
    termbf = temp;
  }
  return 1.0;   /* Non converge ... */
}
