/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

                      File: ststuff.c

 $Id: ststuff.c,v 1.2 2004/10/07 09:00:20 loreti Exp $


 *------------------------------------------------------------------*/

#include <stdio.h>                /* Standard library */
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "ststuff.h"              /* La mia roba */
#include "statfun.h"              /* Da "Numerical recipes in C" */

#ifdef HAVE_READLINE
  #include <readline/readline.h>
  #include <readline/history.h>
#endif /* HAVE_READLINE */

#ifdef HAVE_GSL
  #include <gsl/gsl_cdf.h>
  #include <gsl/gsl_randist.h>
#endif /* HAVE_GSL */

/**
 | Prototipi delle funzioni locali
**/

static void        iner(void);
static const char *getinput(const char *, char *, size_t);

#ifndef HAVE_GSL
static double mypow(double, int);
#endif /* HAVE_GSL */

/**
 | Variabili globali (dichiarate in main)
**/

extern char buffer[];

void bernoulli(void)
{

  /**
   | Distribuzione binomiale
   | =======================
   |
   | I coefficienti binomiali sono calcolati ricorsivamente, usando la
   | relazione C(N,K) = C(N,K-1) * (N-K+1) / K .  Invece i prodotti
   | p^K q^(N-K), anche se potrebbero essere calcolati ricorsivamente,
   | soffrono troppo degli errori di approssimazione e di
   | arrotondamento; per questo vengono ricalcolati per ogni termine.
   |
   | Si richiede che il numero di prove N sia piu' di 2, e che la
   | probabilita' p sia valida (maggiore di zero e minore di uno).
  **/

  for (;;) {
    char *pc;
    int   n;

    if (getinput("* N p (o <EOF> per uscire) ? ", buffer,
                 BUF_LEN) == 0) break;

    if ((n = strtol(buffer, &pc, 10)) > BINOM_LIMIT) {
      printf("\n*** N = %d ?  Troppa grazia...\n", n);
      puts("*** Usa la distribuzione normale o quella "
           "di Poisson!\n");
      continue;
    }

    if (isspace((unsigned char) *pc)   &&   n > 2) {
      double p = strtod(pc, &pc);

      if (LAST_FIELD   &&   p > 0   &&   p < 1) {
        int    i;
        double pcum = 0.0;
        double q    = 1.0 - p;
#ifndef HAVE_GSL
        unsigned long cnk = 1;
#endif /* HAVE_GSL */

        printf("* Distribuzione di Bernoulli: N = %d; p = %g, "
               "q = %g\n", n, p, q);
#ifdef HAVE_GSL
        for (i = 0;   i <= n;   i++) {
          double prob = gsl_ran_binomial_pdf(i, p, n);
# else
        for (i = 0;   i <= n;   ) {
          double prob = cnk * mypow(p, i) * mypow(q, n-i);

          cnk *= n - i;
          cnk /= ++i;
#endif /* HAVE_GSL */
          pcum += prob;
          printf("  i = %d: P =%11.5g (somma =%11.5g)\n",
                 i, prob, pcum);
        }
      } else {
        iner();
      }
    } else {
      iner();
    }
  }

  puts("");
}

void chi(void)
{

  /**
   | Distribuzione del chi-quadro
   | ============================
   |
   | Vengono letti il numero di gradi di liberta' e l'ascissa, e si
   | controlla che entrambi siano positivi; si stampano gli integrali
   | della distribuzione del chi-quadro.
  **/

  for (;;) {
    char   *pc;
    int     ndf;

    if (getinput("* N X (o <EOF>) ? ", buffer, BUF_LEN) == 0) break;

    ndf = strtol(buffer, &pc, 10);

    if (isspace((unsigned char) *pc)   &&   ndf > 0) {
      double x = strtod(pc, &pc);

      if (LAST_FIELD   &&   x > 0) {
#ifdef HAVE_GSL
        double p = gsl_cdf_chisq_P(x, ndf);
#else
        double p = gammp((ndf / 2.0), (x / 2.0));
#endif /* HAVE_GSL */
        double p1 = 1.0 - p;

        printf("* Distribuzione del chi-quadro a %d gradi di "
               "liberta'\n", ndf);
        printf("  Prob(X < %g) =%11.5g    Prob(X > %g) =%11.5g\n",
               x, p, x, p1);
      } else {
        iner();
      }
    } else {
      iner();
    }
  }

  puts("");
}

void fisher(void)
{

  /**
   | Distribuzione di Fisher
   | =======================
   |
   | Vengono letti l'ascissa ed i due gradi di liberta' (che devono
   | essere tutti e tre positivi); si calcola l'integrale della
   | funzione di Fisher.
  **/

  for (;;) {
    char *pc;
    double w;

    if (getinput("* w M N (o <EOF>) ? ", buffer, BUF_LEN) == 0) break;

    w = strtod(buffer, &pc);

    if (isspace((unsigned char) *pc)   &&   w > 0.0) {
      int m = strtol(pc, &pc, 10);

      if (isspace((unsigned char) *pc)   &&   m > 0) {
        int n = strtol(pc, &pc, 10);

        if (LAST_FIELD   &&   n > 0) {
#ifdef HAVE_GSL
          double p = gsl_cdf_fdist_Q(w, m, n);
#else
          double p = betai((n/2.0), (m/2.0), (n/(n+m*w)));
#endif /* HAVE_GSL */
          double p1 = 1.0 - p;

          printf("* Distribuzione di Fisher a %d e %d gradi di "
                 "liberta':\n", m, n);
          printf("Prob(w < %g) =%11.5g     Prob(w > %g) =%11.5g\n",
                 w, p1, w, p);
        } else {
          iner();
        }
      } else {
        iner();
      }
    } else {
      iner();
    }
  }

  puts("");
}

void gauss(void)
{

  /**
   | Distribuzione normale
   | =====================
   |
   | Vengono presentate tre alternative: gli integrali per valori
   | inferiori o superiori ad uno dato, e l'integrale all'interno di
   | un intervallo finito (simmetrico o no rispetto all'origine).
  **/

  for (;;) {
    char   *pc;
    int     che_fare, ancora = 1;
    double  mu, sigma;

    puts("\n\t1 - Probabilita' per N(mu, sigma) di -t < x < +t ;");
    puts("\t2 - Probabilita' per N(mu, sigma) di t1 < x < t2 ;");
    puts("\t3 - Probabilita' per N(mu, sigma) di x < t e x > t ;");
    puts("\t    <EOF> per uscire.\n");

    if ((che_fare = readint("* ? ", 3)) == 0) break;

    if (getinput("* mu sigma (o <EOF>) ? ", buffer, BUF_LEN) == 0)
      break;

    mu = strtod(buffer, &pc);

    if (isspace((unsigned char) *pc)) {
      sigma = strtod(pc, &pc);
      if (!LAST_FIELD   ||   sigma < 0) {
        iner();
        break;
      }
    } else {
      iner();
      break;
    }

    while (ancora) {
      double t, t1, norm, norm1, prob, p1;

      switch (che_fare) {
        case 1:
          if (getinput("* t (o <EOF>) ? ", buffer, BUF_LEN) == 0) {
            ancora = 0;
          } else {
            t = strtod(buffer, &pc);

            if (LAST_FIELD) {
              t    = fabs(t);
              norm = (t - mu)/sigma;
              norm = fabs(norm);
#ifdef HAVE_GSL
              prob  = gsl_cdf_ugaussian_P( norm);
              prob -= gsl_cdf_ugaussian_P(-norm);
#else
              prob = erf(0.7071067814 * norm);
#endif /* HAVE_GSL */
              p1 = 1.0 - prob;
              printf("Prob(-%g < N(%g,%g) < +%g) =%11.5g = 1 -%11.5g\n",
                     t, mu, sigma, t, prob, p1);
            } else {
              iner();
            }
          }
          break;

        case 2:
          if (getinput("* t1 t2 (o <EOF>) ? ", buffer, BUF_LEN) == 0) {
            ancora = 0;
          } else {
            t = strtod(buffer, &pc);

            if (isspace((unsigned char) *pc)) {
              t1 = strtod(pc, &pc);

              if (LAST_FIELD   &&   t1 > t) {
#ifndef HAVE_GSL
                double p2;
#endif /* HAVE_GSL */
                norm  = (t  - mu)/sigma;
                norm1 = (t1 - mu)/sigma;
#ifdef HAVE_GSL
                prob  = gsl_cdf_ugaussian_P(norm);
                prob -= gsl_cdf_ugaussian_P(norm1);
#else
                p1   = erf(0.7071067814 * norm);
                p2   = erf(0.7071067814 * norm1);
                prob = 0.5 * (p2 - p1);
#endif /* HAVE_GSL */
                prob = fabs(prob);
                p1   = 1.0 - prob;
                printf("Prob(%g < N(%g,%g) < %g) =%11.5g = 1 -%11.5g\n",
                       t, mu, sigma, t1, prob, p1);
              } else {
                iner();
              }
            } else {
              iner();
            }
          }
          break;

        case 3:
          if (getinput("* t (o <EOF>) ? ", buffer, BUF_LEN) == 0) {
            ancora = 0;
          } else {
            t = strtod(buffer, &pc);

            if (LAST_FIELD) {
              norm = (t - mu)/sigma;
#ifdef HAVE_GSL
              prob = gsl_cdf_ugaussian_P(norm);
#else
              prob = 0.5 + 0.5 * erf(0.7071067814 * norm);
#endif /* HAVE_GSL */
              p1 = 1.0 - prob;
              printf("Prob(N(%g,%g) < %g) =%11.5g\n", mu, sigma,
                     t, prob);
              printf("Prob(N(%g,%g) > %g) =%11.5g\n", mu, sigma,
                     t, p1);
            } else {
              iner();
            }
          }
          break;
      }
    }
    puts("");
  }
  puts("");
}

void kolsmi(void)
{

  /**
   | Test di Kolmogorov e Smirnov
   |
   | Due scelte: il confronto tra un campione e una distribuzione, o
   | quello tra due campioni.
  **/

  for (;;) {
    int i, e_vai = 1;

    puts("\n\t1 - Confronto tra un campione ed una distribuzione;");
    puts("\t2 - Confronto tra due distribuzioni;");
    puts("\t    <EOF> per uscire.\n");

    if ((i = readint("* ? ", 2)) == 0) break;

    while (e_vai) {
      char   *pc;
      int     n1, n2;
      double  delta, n, p, p1;

      switch (i) {
        case 1:
          if (getinput("* N delta (or <EOF>) ? ", buffer,
                       BUF_LEN) == 0) {
            e_vai = 0;
          } else {
            n1 = strtol(buffer, &pc, 10);

            if (isspace((unsigned char) *pc)   &&   n1 > 1) {
              delta = strtod(pc, &pc);

              if (LAST_FIELD   &&   delta > 0) {
                n  = sqrt(n1);
                n  = (n + 0.12 + 0.11/n) * delta;
                p  = probks(n);
                p1 = 1.0 - p;
                printf("* Test di Kolmogorov & Smirnov, N = %d :\n",
                       n1);
                printf("  Prob(d > %g) = %10.5g = 1 - %10.5g\n",
                       delta, p, p1);
              } else {
                iner();
              }
            } else {
              iner();
            }
          }
          break;

        case 2:
          if (getinput("* N1 N2 delta (or <EOF>) ? ",buffer,
                       BUF_LEN) == 0) {
            e_vai = 0;
          } else {
            n1 = strtol(buffer, &pc, 10);

            if (isspace((unsigned char) *pc)   &&   n1 > 1) {
              n2 = strtol(pc, &pc, 10);

              if (isspace((unsigned char) *pc)   &&   n2 > 1) {
                delta = strtod(pc, &pc);

                if (LAST_FIELD   &&   delta > 0) {
                  n  = n1;
                  n  = sqrt(n * n2 /(n + n2));
                  n  = (n + 0.12 + 0.11/n) * delta;
                  p  = probks(n);
                  p1 = 1.0 - p;
                  printf("* Test di Kolmogorov & Smirnov, N1 = %d, "
                         "N2 = %d :\n", n1, n2);
                  printf("  Prob(d > %g) = %10.5g = 1 - %10.5g\n",
                         delta, p, p1);
                } else {
                  iner();
                }
              } else {
                iner();
              }
            } else {
              iner();
            }
          }
          break;
      }
    }
    puts("");
  }
}

void poisson(void)
{

  /**
   | Distribuzione di Poisson
   | ========================
   |
   | Vengono presentate due scelte: o la probabilita' di ottenere un
   | valore inferiore o superiore ad uno dato; o il calcolo
   | dell'intera distribuzione.  Nel secondo caso, la probabilita' e'
   | calcolata ricorsivamente --- fino a che la distribuzione
   | cumulativa non supera un valore prefissato (POISSON_LIMIT).
  **/

  for (;;) {
    char   *pc;
    int     i, n;
    double  alfa;

    puts("\n\t1 - Probabilita' di valori inferiori o non inferiori "
         "a N;");
    puts("\t2 - Distribuzione completa;");
    puts("\t    <EOF> per uscire.\n");

    if ((i = readint("* ? ", 2)) == 0) break;

    switch (i) {
      case 1:
        if (getinput("* alfa, N ? ", buffer, BUF_LEN) == 0) break;

        alfa = strtod(buffer, &pc);

        if (isspace((unsigned char) *pc)   &&   alfa > 0.0) {
          n = strtol(pc, &pc, 10);

          if (LAST_FIELD   &&   n > 1) {
#ifdef HAVE_GSL
            int    k;
            double p = 0.0, p1;

            for (k = 0;   k < n;   k++) {
              p += gsl_ran_poisson_pdf(k, alfa);
            }
#else
            double p = gammq(n, alfa), p1;
#endif /* HAVE_GSL */
            p1 = 1.0 - p;

            printf("* Distribuzione di Poisson, media = %g:\n",
                   alfa);
            printf("  Prob(x < %d) =%11.5g    Prob(x >= %d) =%11.5g\n",
                   n, p, n, p1);
          } else {
            iner();
          }
        } else {
          iner();
        }
        break;

      case 2:
        if (getinput("* alfa ? ", buffer, BUF_LEN) == 0) break;

        alfa = strtod(buffer, &pc);

        if (LAST_FIELD   &&   alfa > 0.0) {
          double p, psum;

          printf("* Distribuzione di Poisson, media = %g:\n",
                 alfa);
          n = 0;

#ifdef HAVE_GSL
          psum = 0.0;
          for (;;) {
            double p1;

            p = gsl_ran_poisson_pdf(n, alfa);
            psum += p;
            if (psum > POISSON_LIMIT) break;
            p1    = 1.0 - psum;
            printf(" n = %d: P =%11.5g    Somma =%11.5g  = 1 -%11.5g\n",
                   n++, p, psum, p1);
          }
#else
          psum = p = exp(-alfa);
          while (psum < POISSON_LIMIT) {
            double p1 = 1.0 - psum;
            printf(" n = %d: P =%11.5g    Somma =%11.5g  = 1 -%11.5g\n",
                   n, p, psum, p1);
            p    *= alfa;
            p    /= ++n;
            psum += p;
          }
#endif /* HAVE_GSL */
        } else {
          iner();
        }
        break;
    }
  }
  puts("");
}

void student(void)
{

  /**
   | Distribuzione di Student
   | ========================
   |
   | Si calcola l'integrale della distribuzione di Student, dopo aver
   | controllato che il numero di gradi di liberta' sia positivo.
  **/

  for (;;) {
    char *pc;
    int   ndf;

    if (getinput("* N t (o <EOF>) ? ", buffer, BUF_LEN) == 0) break;

    ndf = strtol(buffer, &pc, 10);
    if (isspace((unsigned char) *pc)   &&   ndf > 0) {
      double t = strtod(pc, &pc);

      if (LAST_FIELD) {
        double p, pp, p1, pp1;
#ifndef HAVE_GSL
        double x;
#endif /* HAVE_GSL */

        t   = fabs(t);
#ifdef HAVE_GSL
        p  = gsl_cdf_tdist_P(t, ndf);
        pp = p - gsl_cdf_tdist_P(-t, ndf);
#else
        x  = ndf / (ndf + t * t);
        p  = 1 - 0.5 * betai(0.5*ndf, 0.5, x);
        pp = 2.0 * p - 1.0;
#endif /* HAVE_GSL */
        p1  = (1.0 - p);
        pp1 = (1.0 - pp);

        printf("* Distribuzione di Student a %d gradi di liberta':\n",
               ndf);
        printf("  Prob(t < %g) =%11.5g\n",
               t, p);
        printf("  Prob(t > %g) =%11.5g\n",
               t, p1);
        printf("  Prob(-%g < t < +%f) =%11.5g = 1 -%11.5g\n",
               t, t, pp, pp1);
      } else {
        iner();
      }
    } else {
      iner();
    }
  }

  puts("");
}

static void iner(void)
{

  /**
   | Una versione alquanto semplice di gestione degli errori.
   | Sicuramente migliorabile: si potrebbe ad esempio stampare il tipo
   | di errore e/o il carattere illegale...
  **/

  puts("Errore !!!");
}

#ifndef HAVE_GSL
static double mypow(
  double base,
  int    expon
){

  /**
   | La pow(double base, double expon) della libreria matematica deve
   | essere generale e non puo' quindi essere molto furba; qui si
   | tenta di sveltire i calcoli (entro certi limiti).
  **/

  double temp;

  switch (expon) {
    case 0:                     return 1.0;
    case 1:                     return base;
    case 2:                     return base*base;
    case 3:                     return base*base*base;
    case 4:  temp = base*base;  return temp*temp;
    case 5:  temp = base*base;  return temp*temp*base;
    case 6:  temp = base*base;  return temp*temp*temp;
    default: return pow(base, expon);
  }
}
#endif /* HAVE_GSL */

int readint(
  const char *prompt,
  int         hi
){

  /**
   | Continua a scrivere "prompt" su stdout ed a leggere da stdin fino
   | a che non viene inserito o l'end-of-file, o un numero intero
   | positivo non superiore a "max"; restituisce il numero letto,
   | oppure 0 per l'end-of-file .
  **/

  for (;;) {
    long int  answer;
    char     *pc;

    if (getinput(prompt, buffer, BUF_LEN) == 0) return 0;

    answer = strtol(buffer, &pc, 10);
    if (!LAST_FIELD   ||   answer < 1   ||   answer > hi) {
      iner();
    } else {
      return answer;
    }
  }
}

static const char *getinput(
  const char *prompt,
  char       *buffer,
  size_t      size
) {

  /**
   | "getinput" e' un sostituto di fgets(3) che, se possibile, usa la
   | libreria GNU "readline" (per poter editare i dati mentre li
   | stiamo immettendo, per usare i tasti con le frecce per richiamare
   | le linee precedenti, e cosi' via); *NON* usa, invece, la libreria
   | (collegata a "readline") GNU "history".
   |
   | "getinput" scrive "prompt" sullo stream standard di output, a
   | meno che "prompt" sia un puntatore nullo o che punti a una string
   | vuota (""); poi legge dallo stream standard di input un massimo
   | di "size - 1" caratteri, scrivendoli nel vettore "buffer" e
   | terminandoli con un carattere nullo ('\0').
   |
   | "getinput" restituisce "buffer" in caso di successo; e un
   | puntatore nullo alla lettura dell'end-of-file, oppure in caso di
   | errore in lettura.
  **/

#ifdef HAVE_READLINE
  char *line;
  char *from;

  if ((line = readline(prompt)) == 0) return 0;

  /**
   | La linea letta (se non e' vuota) viene passata all'history ring;
   | e fino a "size" caratteri (compreso il terminatore '\0') vengono
   | spostati in "buffer".
  **/

    if (*line) add_history(line);

    from = line;
    while (--size  &&  (*buffer++ = *from++)) { }
    *buffer = '\0';

  /**
   | Il buffer temporaneo allocato da "readline" viene rilasciato, e
   | si esce.
  **/

  free(line);
  return buffer;
#else
  if (prompt   &&   *prompt) printf("%s", prompt);
  return fgets(buffer, size, stdin);
#endif /* HAVE_READLINE */
}
